{Quantum state reconstruction from homodyne tomography data using the quantum
expectation maximization algorithm.
Written by Alexander Lvovsky, University of Calgary.
See README.TXT or http://iqis.org/quantech/quantechi.php for further details.}

unit MaxLikUnit;

interface

uses
  Math, Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Db, DBTables, DBCGrids, Grids;

type
  TFMaxLik = class(TForm)
    Go: TButton;
    Quit: TButton;
    rhoOut: TStringGrid;
    Save: TButton;
    ldataout: TEdit;
    Label2: TLabel;
    LikLogOut: TEdit;
    Label3: TLabel;
    Load: TButton;
    ItOut: TEdit;
    Label5: TLabel;
    NEdit: TEdit;
    Label6: TLabel;
    AlphaEdit: TEdit;
    Algorithm: TComboBox;
    OpenDialog1: TOpenDialog;
    saveRdata: TCheckBox;
    NoPhase: TCheckBox;
    Stop: TButton;
    EffCorr: TCheckBox;
    EffEdit: TEdit;
    Label1: TLabel;
    Label4: TLabel;
    Label9: TLabel;
    Label8: TLabel;
    Label10: TLabel;
    Label11: TLabel;
    Label12: TLabel;
    Label13: TLabel;
    Label14: TLabel;
    Label15: TLabel;
    Label16: TLabel;
    Label17: TLabel;
    Label18: TLabel;
    Label19: TLabel;
    Label7: TLabel;
    SaveDialog1: TSaveDialog;
    procedure EffCorrClick(Sender: TObject);
    procedure GoClick(Sender: TObject);
    procedure QuitClick(Sender: TObject);
    procedure SaveClick(Sender: TObject);
    procedure LoadClick(Sender: TObject);
    procedure NoPhaseClick(Sender: TObject);
    procedure StopClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    procedure rhoCalc;
    function Delta(i,j:integer):integer;
    function factor(fact:integer):Extended;
    function bincoef(binbb,binaa:Integer):Extended;
    { Private declarations }
  public
    constructor Create(AOwner: TComponent); override;
    { Public declarations }
  end;

const
  maxphotons=15;     {Maximum size of the reconstructed state density matrix}
  maxld=200000;      {Maximum data set length}

{Set these parameters according to your application requirements and memory capacity}

var
  FMaxLik: TFMaxLik;
  data,phase:array[1..maxld] of single;
  data1:array[1..maxld] of integer;
  proj,proj0:array[1..maxld,0..maxphotons,0..maxphotons,1..2] of single;
  rho,R,MX1,MX2,R1R1:array[0..maxphotons,0..maxphotons,1..2] of double;
  B:array[0..maxphotons,0..maxphotons] of double;
  psi:array[-400..400,0..100] of single;
  f,fpsi,frho,fR,fdefault:text;
  defdir,progdir,fn,st,outrhodata,outRdata:string;
  n,i,j,k,l,a,iter,it,ldata:integer;
  LikLog,eta,trace,prob,alpha:double;
  StopPressed:boolean;



implementation

constructor TFMaxLik.Create(AOwner: TComponent);
 var i,j:integer;
 begin

  inherited create(AOwner);
  Algorithm.ItemIndex:=0;

  progdir:=GetCurrentDir;
  try
   assignfile(fpsi,progdir+'\psifunc100.txt');
   reset(fpsi)
  except MessageDlg('Error opening psi-function file',mtError, [mbOk], 0); Application.Terminate
  end;
  for i:=0 to 100 do for j:=-400 to 400 do
    try
      readln(fpsi,psi[j][i]);
     except MessageDlg('Error reading psi-function file',mtError, [mbOk], 0);   Exit
    end;
   closefile(fpsi);


  try
   assignfile(fdefault,progdir+'\defaults.txt');
   reset(fdefault);
   readln(fdefault,defdir);
   closefile(fdefault)
  except defdir:=progdir;
  end;
  OpenDialog1.InitialDir:=defdir;



 end;

function TFMaxLik.Delta(i,j:integer):integer;
 begin if i=j then Delta:=1 else Delta:=0 end;

procedure TFMaxLik.EffCorrClick(Sender: TObject);
begin
  If EffCorr.Checked then EffEdit.enabled:=True else EffEdit.enabled:=False 
end;

procedure TFMaxLik.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  try
   rewrite(fdefault);
   writeln(fdefault,GetCurrentDir);
   closefile(fdefault)
  except MessageDlg('Error saving default directory',mtError, [mbOk], 0);
  end;
end;

procedure TFMaxLik.rhoCalc;
 begin
 StopPressed:=False;
 alpha:=strtofloat(AlphaEdit.Text);
 Go.Enabled:=False;
 Load.Enabled:=False;
 repeat
  Application.ProcessMessages;
  inc(it);
  LikLog:=0;
  for j:=0 to maxphotons do for i:=0 to maxphotons do begin
    r[i][j][1]:=0; r[i][j][2]:=0;
    mx1[i][j][1]:=0; mx1[i][j][2]:=0;
    mx2[i][j][1]:=0; mx2[i][j][2]:=0;
  end;

{Calculate R matrix}
  if NoPhase.Checked then begin
    for k:=1 to ldata do if (data1[k]>=-400) and (data1[k]<=400) then begin
     prob:=0;
     for j:=0 to n do prob:=prob+rho[j][j][1]*proj[k][j][j][1];
     LikLog:=LikLog+ln(prob);
     for j:=0 to n do r[j][j][1]:=r[j][j][1]+proj[k][j][j][1]/prob/ldata;
    end;
    for j:=0 to n do mx2[j][j][1]:=r[j][j][1]*rho[j][j][1];
  end
  else begin
    for k:=1 to ldata do if (data1[k]>=-400) and (data1[k]<=400) then begin
     prob:=0;
     for j:=0 to n do for i:=0 to n do
      prob:=prob+rho[j][i][1]*proj[k][i][j][1]-rho[j][i][2]*proj[k][i][j][2];
     LikLog:=LikLog+ln(prob);
     for j:=0 to n do for i:=0 to n do begin
      r[i][j][1]:=r[i][j][1]+proj[k][i][j][1]/prob/ldata;
      r[i][j][2]:=r[i][j][2]+proj[k][i][j][2]/prob/ldata;
     end;
    end;

    if Algorithm.ItemIndex=1 then
     for j:=0 to n do for i:=0 to n do begin
      r[i][j][1]:=alpha*r[i][j][1]+Delta(i,j);
      r[i][j][2]:=alpha*r[i][j][2];
     end;

{Multiply R by the density matrix on both sides}
    for j:=0 to n do for i:=0 to n do for a:=0 to n do begin
     mx1[j][i][1]:=mx1[j][i][1]+r[j][a][1]*rho[a][i][1]-r[j][a][2]*rho[a][i][2];
     mx1[j][i][2]:=mx1[j][i][2]+r[j][a][1]*rho[a][i][2]+r[j][a][2]*rho[a][i][1]
    end;

    for j:=0 to n do for i:=0 to n do for a:=0 to n do begin
      mx2[j][i][1]:=mx2[j][i][1]+mx1[j][a][1]*r[a][i][1]-mx1[j][a][2]*r[a][i][2];
      mx2[j][i][2]:=mx2[j][i][2]+mx1[j][a][1]*r[a][i][2]+mx1[j][a][2]*r[a][i][1]
    end;
  end;

{Normalize}
  trace:=0; for j:=0 to n do trace:=trace+mx2[j][j][1];
  for j:=0 to n do for i:=0 to n do begin
   rho[j][i][1]:=mx2[j][i][1]/trace;
   rho[j][i][2]:=mx2[j][i][2]/trace;
  end;

  ItOut.Text:=inttostr(it);
  LikLogOut.text:=floattostrF(LikLog,ffgeneral,15,0);
  for i:=0 to n do for j:=0 to n do rhoOut.Cells[i+1,j+1]:=
     floattostrF(rho[j][i][1],ffgeneral,5,0)+'+'+floattostrF(rho[j][i][2],ffgeneral,5,0)+'i';
  FMaxLik.repaint
 until StopPressed;
 Go.Enabled:=True;
 Load.Enabled:=True;
 end;


{$R *.DFM}

procedure TFMaxLik.GoClick(Sender: TObject);
begin
   rhoCalc;
end;

procedure TFMaxLik.QuitClick(Sender: TObject);
begin
  close
end;

procedure TFMaxLik.SaveClick(Sender: TObject);
begin
 SaveDialog1.FileName:=ChangeFileExt(OpenDialog1.FileName,'_rho.txt');
 if SaveDialog1.Execute then outrhodata:=SaveDialog1.FileName else exit;
 try
  assignfile(frho,outrhodata); rewrite(frho);
   for j:=0 to n do begin
    for i:=0 to n do begin
     st:=floattostrF(rho[j][i][1],ffgeneral,9,0);
     st:=st+' '+floattostrF(rho[j][i][2],ffgeneral,9,0);
     if i<n then write(frho,st+'    ') else write(frho,st) ;
    end; if j<n then writeln(frho,' ')
   end;
   closefile(frho);
 except MessageDlg('Error saving file',mtError, [mbOk], 0); exit
 end;

 if saveRdata.checked then begin
    SaveDialog1.FileName:=ChangeFileExt(OpenDialog1.FileName,'.R');
    if SaveDialog1.Execute then outRdata:=SaveDialog1.FileName else exit;
    try
      assignfile(fR,outRdata); rewrite(fR);
      for j:=0 to n do begin
       for i:=0 to n do begin
        st:=floattostrF(r[j][i][1],ffgeneral,5,0);
        st:=st+' '+floattostrF(r[j][i][2],ffgeneral,5,0);
        if i<n then write(fr,st+'    ') else write(fr,st) ;
       end; if j<n then writeln(fr,' ')
      end;
      closefile(fr);
    except MessageDlg('Error saving R file',mtError, [mbOk], 0); exit
    end;
 end;
end;

procedure TFMaxLik.StopClick(Sender: TObject);
begin
  StopPressed:=True;
end;

procedure TFMaxLik.LoadClick(Sender: TObject);
begin
  Go.Enabled:=False;
  it:=0;
  if EffCorr.Checked then eta:=StrToFloat(EffEdit.Text) else eta:=1;
  n:=strtoint(NEdit.text);
  if n>maxphotons
   then begin MessageDlg('Maximum allowed photon number is '+Inttostr(maxphotons),mtError, [mbOk], 0); Exit end;
  for k:=1 to ldata do for j:=0 to n do for i:=0 to n do begin
    proj[k][i][j][1]:=0;
    proj[k][i][j][2]:=0;
  end;
  for j:=0 to n do for i:=0 to n do
    if j>=i then B[j][i]:=sqrt(bincoef(j,i)*power(eta,i)*power((1-eta),j-i)) else B[j][i]:=0;

  for j:=0 to n do for i:=0 to n do if i=j then rho[i][j][1]:=1/(n+1) else  rho[j][i][1]:=0;
  for j:=0 to n do for i:=0 to n do rho[j][i][2]:=0;

  for i:=0 to 17 do for j:=0 to 17 do rhoOut.Cells[i,j]:='';
  for i:=0 to n do begin
   rhoOut.Cells[0,i+1]:=inttostr(i); rhoOut.Cells[i+1,0]:=inttostr(i);
   for j:=0 to n do rhoOut.Cells[i+1,j+1]:=
     floattostrF(rho[j][i][1],ffgeneral,5,0)+'+'+floattostrF(rho[j][i][2],ffgeneral,5,0)+'i';
  end;

  OpenDialog1.Title:='Input quadrature file';
  if OpenDialog1.Execute then begin
   defdir:=GetCurrentDir;
   fn:=OpenDialog1.FileName;
   try
    assignfile(f,fn);  reset(f)
   except MessageDlg('Error opening input data file',mtError, [mbOk], 0); Exit
   end;
   i:=0;
   ItOut.Text:='';
   LikLogOut.Text:='';
   repeat
    inc(i);
    if i>maxld
     then begin MessageDlg('Maximum allowed data set length is '+Inttostr(maxld),mtError, [mbOk], 0); Exit end;
    if NoPhase.Checked then readln(f,data[i]) else readln(f,phase[i],data[i])
   until eof(f);
   closefile(f);
  end;
  ldata:=i;
  ldataout.text:= inttostr(ldata);
  for i:=1 to ldata do data1[i]:=round(data[i]*100);

  if NoPhase.Checked then begin
    for k:=1 to ldata do {if k mod 4 = 0 then} if (data1[k]>=-400) and (data1[k]<=400) then begin
     for j:=0 to n do for i:=0 to n do begin
       proj0[k][i][j][1]:=psi[data1[k]][i]*psi[data1[k]][i]*delta(i,j);
       proj0[k][i][j][2]:=0;
     end;
    end
  end else begin
    for k:=1 to ldata do {if k mod 4 = 0 then} if (data1[k]>=-400) and (data1[k]<=400) then begin
     for j:=0 to n do for i:=0 to n do begin
       proj0[k][i][j][1]:=psi[data1[k]][i]*psi[data1[k]][j]*cos((j-i)*phase[k]);
       proj0[k][i][j][2]:=psi[data1[k]][i]*psi[data1[k]][j]*sin((j-i)*phase[k]);
     end;
    end;
  end;
  if EffCorr.Checked then
    for k:=1 to ldata do
     for j:=0 to n do for i:=0 to n do for l:=0 to min(i,j) do begin
       proj[k][i][j][1]:=proj[k][i][j][1]+proj0[k][i-l][j-l][1]*B[i][i-l]*B[j][j-l];
       proj[k][i][j][2]:=proj[k][i][j][2]+proj0[k][i-l][j-l][2]*B[i][i-l]*B[j][j-l];
     end
    else for k:=1 to ldata do for j:=0 to n do for i:=0 to n do for l:=1 to 2 do
      proj[k][i][j][l]:=proj0[k][i][j][l];
  Go.Enabled:=True;
end;

procedure TFMaxLik.NoPhaseClick(Sender: TObject);
begin
  if NoPhase.Checked then begin
    Algorithm.enabled:=False;
    AlphaEdit.enabled:=False;
  end else begin
    Algorithm.enabled:=True;
    AlphaEdit.enabled:=True;
  end;
end;

function TFMaxLik.factor(fact:integer):Extended;
var i:integer;
begin
     factor:=1;
     for i:=1 to fact do factor:=Result*i;
end;

function TFMaxLik.bincoef(binbb,binaa:Integer):Extended;
begin
  if binaa > binbb
      then bincoef:=0
      else if binaa=binbb
          then bincoef:=1
          else bincoef:=factor(binbb)/(factor(binaa)*factor(binbb-binaa));
end;



end.
